==> VERSION_2_4_24.tar.gz: duma/sem_inc.c <==

/*
 * DUMA - Red-Zone memory allocator.
 * Copyright (C) 2002-2005 Hayati Ayguen <h_ayguen@web.de>, Procitec GmbH
 * License: GNU LGPL (GNU Lesser General Public License, see COPYING-GPL)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * FILE CONTENTS:
 * internal implementation file
 * contains thread safety functions (semaphore lock/release)
 */

#include "duma_config.h"
#include "sem_inc.h"
#include "print.h"

#ifndef DUMA_NO_THREAD_SAFETY

#ifndef WIN32
  #include <pthread.h>
  #include <semaphore.h>
#else
  #define WIN32_LEAN_AND_MEAN 1
  #include <windows.h>
  #include <winbase.h>
#endif


/*
 * DUMA_sem is a semaphore used to allow one thread at a time into
 * these routines.
 * Also, we use semInited as a boolean to see if we should be
 * using the semaphore.
 * semThread is set to the thread id of the thread that currently
 * has the semaphore so that when/if it tries to get the semaphore
 * again (realloc calling malloc/free) - nothing will happen to the
 * semaphore.
 * semDepth is used to keep track of how many times the same thread
 * gets the semaphore - so we know when it is actually freed.
 */



#ifndef WIN32
static sem_t      DUMA_sem = { 0 };
static pthread_t  semThread = (pthread_t) 0;

#else /* WIN32 */
static SECURITY_ATTRIBUTES  semSecAttr;

#ifndef UNICODE
#define SEM_NAME_TYPE char
#define SEM_STRCPY    strcpy
#define SEM_STRCAT    strcat
static char       semObjectName[] = "EFence_";
#else
#define SEM_NAME_TYPE wchar_t
#define SEM_STRCPY    wcscpy
#define SEM_STRCAT    wcscat
static wchar_t    semObjectName[] = L"EFence_";
#endif

static DWORD      semThread = 0;
static HANDLE     semHandle = 0;
#endif

static int        semInited = 0;
static int        semDepth  = 0;


void
#ifdef DUMA_GNU_INIT_ATTR
__attribute ((constructor))
#endif
DUMA_init_sem(void)
{
#ifndef WIN32
  if (semInited)
    return;

  if (sem_init(&DUMA_sem, 0, 1) >= 0)
    semInited = 1;
#else
  SEM_NAME_TYPE   semLocalName[32];
  SEM_NAME_TYPE   acPID[16];
  DWORD pid;

  if (semInited)
    return;

  pid = GetCurrentProcessId();
  SEM_STRCPY(semLocalName, semObjectName);
  /* append ProcessId() to get inter-process unique semaphore name */
  acPID[0] = 'A' + (SEM_NAME_TYPE)( (pid >> 28) & 0x0F );
  acPID[1] = 'A' + (SEM_NAME_TYPE)( (pid >> 24) & 0x0F );
  acPID[2] = 'A' + (SEM_NAME_TYPE)( (pid >> 20) & 0x0F );
  acPID[3] = 'A' + (SEM_NAME_TYPE)( (pid >> 16) & 0x0F );
  acPID[4] = 'A' + (SEM_NAME_TYPE)( (pid >> 12) & 0x0F );
  acPID[5] = 'A' + (SEM_NAME_TYPE)( (pid >>  8) & 0x0F );
  acPID[6] = 'A' + (SEM_NAME_TYPE)( (pid >>  4) & 0x0F );
  acPID[7] = 'A' + (SEM_NAME_TYPE)( (pid      ) & 0x0F );
  acPID[8] = 0;
  SEM_STRCAT( semLocalName, acPID );

  semSecAttr.nLength = sizeof(semSecAttr);
  semSecAttr.lpSecurityDescriptor = NULL;
  semSecAttr.bInheritHandle = FALSE;

  semHandle = CreateSemaphore( &semSecAttr      /* pointer to security attributes */
                               , 1              /* initial count */
                               , 10             /* maximum count */
                               , semLocalName   /* pointer to semaphore-object name */
                             );
  semInited = 1;
#endif
  if (!semInited)     DUMA_Abort("\nCouldn't initialise semaphore");
}

#ifndef WIN32
# define DUMA_thread_self() pthread_self()
#else
# define DUMA_thread_self() GetCurrentThreadId()
#endif


void DUMA_get_sem(void)
{
#ifndef DUMA_GNU_INIT_ATTR
  if (!semInited)     DUMA_init_sem();    /* initialize if necessary */
#endif

  if (semThread != DUMA_thread_self())
  {
#ifndef WIN32
    while (sem_wait(&DUMA_sem) < 0);   /* wait for the semaphore. */
#else
    while (WaitForSingleObject(semHandle, 1000) != WAIT_OBJECT_0) ; /* wait for the semaphore. */
#endif
    semThread = DUMA_thread_self();     /* let everyone know who has the semaphore. */
  }
  ++semDepth;                         /* increment semDepth - push one stack level */
}


void DUMA_rel_sem(void)
{
#ifndef DUMA_GNU_INIT_ATTR
  if (!semInited)     DUMA_Abort("\nSemaphore isn't initialised");
#endif

  if (!semThread)     DUMA_Abort("\nSemaphore isn't owned by this thread");
  if (semDepth <= 0)  DUMA_Abort("\nSemaphore isn't locked");

  if (!(--semDepth))              /* decrement semDepth - popping one stack level */
  {
#ifndef WIN32
    semThread = (pthread_t) 0;        /* zero this before actually free'ing the semaphore. */
    if (sem_post(&DUMA_sem) < 0)
#else
    semThread = 0;                    /* zero this before actually free'ing the semaphore. */
    if (0 == ReleaseSemaphore(semHandle, 1 /* amount to add to current count */, NULL) )
#endif
      DUMA_Abort("Failed to post the semaphore.");
  }
}

#else

/* for not having an empty file */
static int dummy = 0;

#endif /* DUMA_NO_THREAD_SAFETY */

==> VERSION_2_4_26.tar.gz: duma/sem_inc.c <==

/*
 * DUMA - Red-Zone memory allocator.
 * Copyright (C) 2002-2005 Hayati Ayguen <h_ayguen@web.de>, Procitec GmbH
 * License: GNU LGPL (GNU Lesser General Public License, see COPYING-GPL)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * FILE CONTENTS:
 * internal implementation file
 * contains thread safety functions (semaphore lock/release)
 */

#include "duma_config.h"
#include "sem_inc.h"
#include "print.h"

#ifndef DUMA_NO_THREAD_SAFETY

#ifndef WIN32
  #include <pthread.h>
  #include <semaphore.h>
#else
  #define WIN32_LEAN_AND_MEAN 1
  #include <windows.h>
  #include <winbase.h>
#endif


/*
 * DUMA_sem is a semaphore used to allow one thread at a time into
 * these routines.
 * Also, we use semInited as a boolean to see if we should be
 * using the semaphore.
 * semThread is set to the thread id of the thread that currently
 * has the semaphore so that when/if it tries to get the semaphore
 * again (realloc calling malloc/free) - nothing will happen to the
 * semaphore.
 * semDepth is used to keep track of how many times the same thread
 * gets the semaphore - so we know when it is actually freed.
 */


#ifndef WIN32

  #define DUMA_thread_self()  pthread_self()

  static sem_t      DUMA_sem = { 0 };
  static pthread_t  semThread = (pthread_t) 0;

#else /* WIN32 */

  #define DUMA_thread_self()  GetCurrentThreadId()

  #ifndef UNICODE
    #define SEM_NAME_TYPE char
    #define SEM_STRCPY    strcpy
    #define SEM_STRCAT    strcat
    static char    semObjectName[] = "DUMA_";
  #else
    #define SEM_NAME_TYPE wchar_t
    #define SEM_STRCPY    wcscpy
    #define SEM_STRCAT    wcscat
    static wchar_t semObjectName[] = L"DUMA_";
  #endif

  static SECURITY_ATTRIBUTES  semSecAttr;
  static DWORD      semThread = 0;
  static HANDLE     semHandle = 0;

#endif


static int semInInit = 0;
static int semInited = 0;
static int semDepth  = 0;


void
DUMA_init_sem(void)
{
#ifdef WIN32
  SEM_NAME_TYPE   semLocalName[32];
  SEM_NAME_TYPE   acPID[16];
  DWORD pid;
#endif

  /* avoid recursive call to sem_init(),
   * when sem_init() calls malloc() or other allocation function
   */
  if (semInited || semInInit)
    return;
  semInInit = 1;

#ifndef WIN32
  if (sem_init(&DUMA_sem, 0, 1) >= 0)
    semInited = 1;
#else
  pid = GetCurrentProcessId();
  SEM_STRCPY(semLocalName, semObjectName);
  /* append ProcessId() to get inter-process unique semaphore name */
  acPID[0] = 'A' + (SEM_NAME_TYPE)( (pid >> 28) & 0x0F );
  acPID[1] = 'A' + (SEM_NAME_TYPE)( (pid >> 24) & 0x0F );
  acPID[2] = 'A' + (SEM_NAME_TYPE)( (pid >> 20) & 0x0F );
  acPID[3] = 'A' + (SEM_NAME_TYPE)( (pid >> 16) & 0x0F );
  acPID[4] = 'A' + (SEM_NAME_TYPE)( (pid >> 12) & 0x0F );
  acPID[5] = 'A' + (SEM_NAME_TYPE)( (pid >>  8) & 0x0F );
  acPID[6] = 'A' + (SEM_NAME_TYPE)( (pid >>  4) & 0x0F );
  acPID[7] = 'A' + (SEM_NAME_TYPE)( (pid      ) & 0x0F );
  acPID[8] = 0;
  SEM_STRCAT( semLocalName, acPID );

  semSecAttr.nLength = sizeof(semSecAttr);
  semSecAttr.lpSecurityDescriptor = NULL;
  semSecAttr.bInheritHandle = FALSE;

  semHandle = CreateSemaphore( &semSecAttr      /* pointer to security attributes */
                               , 1              /* initial count */
                               , 10             /* maximum count */
                               , semLocalName   /* pointer to semaphore-object name */
                             );
  semInited = 1;
#endif

  semInInit = 0;

  if (!semInited)     DUMA_Abort("\nCouldn't initialise semaphore");
}


void DUMA_get_sem(void)
{
  if (semInInit)      return;             /* avoid recursion */
  if (!semInited)     DUMA_init_sem();    /* initialize if necessary */

  if (semThread != DUMA_thread_self())
  {
#ifndef WIN32
    while (sem_wait(&DUMA_sem) < 0);   /* wait for the semaphore. */
#else
    while (WaitForSingleObject(semHandle, 1000) != WAIT_OBJECT_0) ; /* wait for the semaphore. */
#endif
    semThread = DUMA_thread_self();     /* let everyone know who has the semaphore. */
  }
  ++semDepth;                         /* increment semDepth - push one stack level */
}


void DUMA_rel_sem(void)
{
  if (semInInit)      return;             /* avoid recursion */
  if (!semInited)     DUMA_Abort("\nSemaphore isn't initialised");

  if (!semThread)     DUMA_Abort("\nSemaphore isn't owned by this thread");
  if (semDepth <= 0)  DUMA_Abort("\nSemaphore isn't locked");

  if (!(--semDepth))              /* decrement semDepth - popping one stack level */
  {
#ifndef WIN32
    semThread = (pthread_t) 0;        /* zero this before actually free'ing the semaphore. */
    if (sem_post(&DUMA_sem) < 0)
#else
    semThread = 0;                    /* zero this before actually free'ing the semaphore. */
    if (0 == ReleaseSemaphore(semHandle, 1 /* amount to add to current count */, NULL) )
#endif
      DUMA_Abort("Failed to post the semaphore.");
  }
}

#else

/* for not having an empty file */
static int dummy = 0;

#endif /* DUMA_NO_THREAD_SAFETY */
